import random
import string


def generate_token(length: int) -> str:
    """Generate random alphanumeric string of the requested lenght."""
    chars = string.ascii_letters + string.digits
    return "".join(random.choice(chars) for _ in range(length))



import argparse
import logging
import os
import time
from configparser import ConfigParser
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List
from urllib.parse import quote_plus, unquote_plus

import multivolumefile
import py7zr

from . import __app_name__, __version__
from .client import ToDusClient

formatter = logging.Formatter("%(levelname)s-%(name)s-%(asctime)s-%(message)s")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler_default = logging.StreamHandler()
handler_default.setFormatter(formatter)
handler_error = logging.FileHandler("logs", encoding="utf-8")
handler_error.setLevel(logging.ERROR)
handler_error.setFormatter(formatter)
logger.addHandler(handler_default)
logger.addHandler(handler_error)

client = ToDusClient()
config = ConfigParser()
MAX_RETRY = 3
DOWN_TIMEOUT = 60 * 20  # 20 MINS


def write_txt(filename: str, urls: List[str], parts: List[str]) -> str:
    txt = "\n".join(f"{down_url}\t{name}" for down_url, name in zip(urls, parts))
    path = Path(f"{filename}.txt").resolve()
    with open(path, "w", encoding="utf-8") as f:
        f.write(txt)
    return str(path)


def split_upload(
    token: str, path: str, part_size: int, max_retry: int = MAX_RETRY
) -> str:
    global client

    with open(path, "rb") as file:
        data = file.read()

    filename = Path(path).name

    logger.info("Compressing parts...")

    with TemporaryDirectory() as tempdir:
        with multivolumefile.open(
            Path(f"{tempdir}/{filename}.7z"),
            "wb",
            volume=part_size,
        ) as vol:
            with py7zr.SevenZipFile(vol, "w") as archive:  # type: ignore
                archive.writestr(data, filename)
        del data
        parts = sorted(_file.name for _file in Path(tempdir).iterdir())
        parts_count = len(parts)

        urls = []

        for i, name in enumerate(parts, 1):
            retry = 0
            up_done = False
            logger.info(f"Uploading {i}/{parts_count}: {name}")

            temp_path = Path(f"{tempdir}/{name}")

            while not up_done and retry < max_retry:
                try:
                    urls.append(client.upload_file(token, temp_path, index=i))
                except Exception as ex:
                    logger.error(ex)
                    retry += 1
                    if retry == max_retry:
                        raise ValueError(
                            f"Failed to upload part {i} ({temp_path.stat().st_size:,}B): {ex}"
                        )
                    logger.info(f"Retrying: {retry}...")
                    time.sleep(15)
                else:
                    up_done = True

        path = write_txt(filename, urls, parts)
    return path


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=__app_name__,
        description="ToDus Client for S3",
    )
    parser.add_argument(
        "-n",
        "--number",
        dest="number",
        metavar="PHONE-NUMBER",
        help="account's phone number",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--config-folder",
        dest="folder",
        type=str,
        default=".",
        help="folder where account configuration will be saved/loaded",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=__version__,
        help="show program's version number and exit.",
    )

    subparsers = parser.add_subparsers(dest="command")

    _ = subparsers.add_parser(name="login", help="authenticate in server")

    up_parser = subparsers.add_parser(name="upload", help="upload file")
    up_parser.add_argument(
        "-p",
        "--part-size",
        dest="part_size",
        type=int,
        default=0,
        help="if given, the file will be split in parts of the given size in bytes",
    )
    up_parser.add_argument("file", nargs="+", help="file to upload")

    down_parser = subparsers.add_parser(name="download", help="download file")
    down_parser.add_argument("url", nargs="+", help="url to download or txt file path")

    return parser


def register(client: ToDusClient, phone: str) -> str:
    client.request_code(phone)
    pin = input("Enter PIN:").strip()
    password = client.validate_code(phone, pin)
    logger.debug("PASSWORD: %s", password)
    return password


def read_config(phone: str, folder: str = ".") -> ConfigParser:
    config_path = Path(folder) / Path(f"{phone}.ini")
    config.read(config_path)
    return config


def save_config(phone: str, folder: str = ".") -> None:
    with open(Path(folder) / Path(f"{phone}.ini"), "w") as configfile:
        config.write(configfile)


def get_default(dtype, dkey: str, phone: str, folder: str, dvalue: str = ""):
    return dtype(
        read_config(phone, folder)["DEFAULT"].get(dkey, str(dvalue))
        if "DEFAULT" in config
        else dvalue
    )


def _upload(token: str, args: argparse.Namespace, max_retry: int):
    for path in args.file:
        filename = Path(path).name
        logger.info(f"Uploading: {filename}")
        if args.part_size:
            txt = split_upload(token, path, args.part_size, max_retry=max_retry)
            logger.info(f"TXT: {txt}")
        else:
            # TODO: Retry for single file
            filename_path = Path(path)
            file_uri = client.upload_file(token, filename_path)
            down_url = f"{file_uri}?name={quote_plus(filename)}"
            logger.info(f"URL: {down_url}")
            txt = write_txt(filename, urls=[file_uri], parts=[filename])
            logger.info(f"TXT: {txt}")


def _download(
    token: str, args: argparse.Namespace, down_timeout: float, max_retry: int
):
    urls = []

    while args.url:
        retry = 0
        down_done = False
        file_uri = args.url.pop(0)

        # Extract URLS from current TXT
        if os.path.exists(file_uri):
            with open(file_uri) as fp:
                for line in fp.readlines():
                    line = line.strip()
                    if line:
                        _url, _filename = line.split(maxsplit=1)
                        urls.append(f"{_url}?name={_filename}")

                args.url = urls + args.url
                continue

        if urls:
            count_files = len(urls)
            plural = "" if count_files <= 1 else "s"
            logger.info(f"Downloading: {count_files} file{plural}")

        logger.info(
            f"Downloading: {file_uri}",
        )
        file_uri, name = file_uri.split("?name=", maxsplit=1)
        name = unquote_plus(name)
        size = 0

        while not down_done and retry < max_retry:
            try:
                size = client.download_file(token, file_uri, name, down_timeout)
            except Exception as ex:
                logger.error(ex)

            if size:
                logger.debug(
                    f"File: {name}, Size: {size // 1024}",
                )
                down_done = True

            if not down_done or not size:
                retry += 1
                if retry == max_retry:
                    break
                logger.info(f"Retrying: {retry}...")
                time.sleep(15)

        urls = []


def main() -> None:
    global client, config, logger

    parser = get_parser()
    args = parser.parse_args()
    folder: str = args.folder
    phone: str = f"535{args.number[-7:]}"  # Country Code + phone
    password: str = get_default(str, "password", phone, folder, "")
    max_retry: int = get_default(int, "max_retry", phone, folder, str(MAX_RETRY))
    config["DEFAULT"]["max_retry"] = str(max_retry)
    down_timeout: float = get_default(
        float, "down_timeout", phone, folder, str(DOWN_TIMEOUT)
    )
    config["DEFAULT"]["down_timeout"] = str(down_timeout)
    production: bool = get_default(bool, "production", phone, folder, "True")
    config["DEFAULT"]["production"] = str(production)

    if production:
        logging.raiseExceptions = False
    else:
        logger.setLevel(logging.DEBUG)

    if not password and args.command != "login":
        print("ERROR: account not authenticated, login first.")
        return

    if args.command == "upload":
        token = client.login(phone, password)
        logger.debug(f"Token: '{token}'")
        _upload(token, args, max_retry)
    elif args.command == "download":
        token = client.login(phone, password)
        logger.debug(f"Token: '{token}'")
        _download(token, args, down_timeout, max_retry)
    elif args.command == "login":
        password = register(client, phone)
        token = client.login(phone, password)
        logger.debug(f"Token: '{token}'")

        config["DEFAULT"]["password"] = password
        config["DEFAULT"]["token"] = token
    else:
        parser.print_usage()

    save_config(phone, folder)



class AuthenticationError(Exception):
    """Failed to authenticate with the server."""


class AbortError(Exception):
    """User canceled the operation."""


class EndOfStreamError(Exception):
    """Unexpected empty response from the xmpp server."""



import json
import re
import socket
import ssl
from base64 import b64decode, b64encode
from typing import Tuple

from .errors import AuthenticationError, EndOfStreamError
from .util import generate_token

_BUFFERSIZE = 1024 * 1024


def _get_socket() -> ssl.SSLSocket:
    socket_ = socket.socket(socket.AF_INET)
    socket_.settimeout(15)
    ssl_socket = ssl.wrap_socket(socket_, ssl_version=ssl.PROTOCOL_TLSv1_2)
    ssl_socket.connect(("im.todus.cu", 1756))
    ssl_socket.send(
        b"<stream:stream xmlns='jc' o='im.todus.cu' xmlns:stream='x1' v='1.0'>"
    )
    return ssl_socket


def _negociate_start(
    response: str, ssl_socket: ssl.SSLSocket, authstr: bytes, sid: str
) -> bool:
    if response.startswith(
        "<?xml version='1.0'?><stream:stream i='"
    ) and response.endswith("xmlns:stream='x1' f='im.todus.cu' xmlns='jc'>"):
        return True

    data: bytes = b""
    auth_stream: str = (
        "<stream:features>"
        "<es xmlns='x2'><e>PLAIN</e><e>X-OAUTH2</e></es>"
        "<register xmlns='http://jabber.org/features/iq-register'/>"
        "</stream:features>"
    )
    if response == auth_stream:
        data = b"<ah xmlns='ah:ns' e='PLAIN'>" + authstr + b"</ah>"
        ssl_socket.send(data)
        return True

    if response == "<ok xmlns='x2'/>":
        data = b"<stream:stream xmlns='jc' o='im.todus.cu' xmlns:stream='x1' v='1.0'>"
        ssl_socket.send(data)
        return True

    if "<stream:features><b1 xmlns='x4'/>" in response:
        data = f"<iq i='{sid}-1' t='set'><b1 xmlns='x4'></b1></iq>".encode("utf-8")
        ssl_socket.send(data)
        return True

    return False


def _parse_token(token: str) -> Tuple[str, bytes]:
    assert token, "Invalid Token 'None'"
    phone = json.loads(b64decode(token.split(".")[1]).decode())["username"]
    authstr = b64encode((chr(0) + phone + chr(0) + token).encode("utf-8"))
    return phone, authstr


def reserve_url(token: str, filesize: int) -> Tuple[str, str]:
    """Reserve file URL to upload.

    Returns a tuple with upload and download URLs.
    """
    data: bytes = b""
    phone, authstr = _parse_token(token)
    sid = generate_token(5)
    ssl_socket = _get_socket()

    while True:
        response = ssl_socket.recv(_BUFFERSIZE).decode()
        if _negociate_start(response, ssl_socket, authstr, sid):
            continue

        if f"t='result' i='{sid}-1'>" in response:
            data = b"<en xmlns='x7' u='true' max='300'/>"
            ssl_socket.send(data)
            data = (
                f"<iq i='{sid}-3' t='get'>"
                f"<query xmlns='todus:purl' type='0' persistent='false'"
                f" size='{filesize}' room=''></query>"
                "</iq>"
            ).encode("utf-8")
            ssl_socket.send(data)
            continue

        if response.startswith("<ed u='true' max='300'"):
            data = f"<p i='{sid}-4'></p>".encode("utf-8")
            ssl_socket.send(data)
            continue

        if response.startswith(f"<iq o='{phone}@im.todus.cu"):
            match = re.match(r".*put='(.*)' get='(.*)' stat.*", response)
            assert match, f"Unexpected response: {response}"
            up_url = match.group(1).replace("amp;", "")
            down_url = match.group(2)
            return (up_url, down_url)

        if "<not-authorized/>" in response:
            raise AuthenticationError()

        if not response:
            raise EndOfStreamError()


def get_real_url(token: str, url: str) -> str:
    """Get authenticated URL."""
    data: bytes = b""
    authstr = _parse_token(token)[1]
    sid = generate_token(5)
    ssl_socket = _get_socket()

    while True:
        response = ssl_socket.recv(_BUFFERSIZE).decode()
        if _negociate_start(response, ssl_socket, authstr, sid):
            continue

        if f"t='result' i='{sid}-1'>" in response:
            data = (
                f"<iq i='{sid}-2' t='get'>"
                f"<query xmlns='todus:gurl' url='{url}'></query>"
                "</iq>"
            ).encode("utf-8")
            ssl_socket.send(data)
            continue

        if f"t='result' i='{sid}-2'>" in response and "status='200'" in response:
            match = re.match(".*du='(.*)' stat.*", response)
            assert match, f"Unexpected response: {response}"
            return match.group(1).replace("amp;", "")

        if "<not-authorized/>" in response:
            raise AuthenticationError()

        if not response:
            raise EndOfStreamError()



import logging
import socket
import string
from pathlib import Path
from queue import Empty, Queue
from typing import Any, Callable, Dict, Optional, Tuple

import requests
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from .errors import AbortError
from .s3 import get_real_url, reserve_url
from .util import generate_token

DEFAULT_TIMEOUT = 60 * 2  # 2 minutes
CHUNK_SIZE = 1024
logger = logging.getLogger(__name__)


class ResultProcess:
    def __init__(self, target, *args, **kwargs) -> None:
        self._real_target = target
        self.args = args
        self.kwargs = kwargs
        self._result_queue: Queue = Queue()
        self._failed = False

    def start(self, *args, **kwargs) -> None:
        try:
            self._result_queue.put(self._real_target(*args, **kwargs))
        except Exception as ex:
            self._failed = True
            self._result_queue.put(ex)

    def abort(self) -> None:
        self._failed = True
        self._result_queue.put(AbortError())

    def get_result(self, timeout: float = None):
        result = None
        try:
            result = self._result_queue.get(timeout=timeout)
        except Empty:
            raise TimeoutError("Operation timed out.")
        if self._failed:
            raise result
        return result


class ToDusClient:
    """Class interact with the Todus API."""

    def __init__(
        self, version_name: str = "0.40.16", version_code: str = "21820"
    ) -> None:
        self.version_name = version_name
        self.version_code = version_code

        self.timeout = DEFAULT_TIMEOUT
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Accept-Encoding": "gzip",
            }
        )
        self._process: Optional[ResultProcess] = None

    def _run_task(self, task: Callable, timeout: float, *args, **kwargs) -> Any:
        self._process = ResultProcess(task, *args, **kwargs)
        try:
            self._process.start(*args, **kwargs)
            return self._process.get_result(timeout)
        except Exception as ex:
            logger.error(ex)
            self.abort()

    def abort(self) -> None:
        if self._process is not None:
            self._process.abort()
            self._process = None
        self.session.close()

    @property
    def auth_ua(self) -> str:
        """User Agent used for authentication."""
        return f"ToDus {self.version_name} Auth"

    @property
    def upload_ua(self) -> str:
        """User Agent used for uploads."""
        return f"ToDus {self.version_name} HTTP-Upload"

    @property
    def download_ua(self) -> str:
        """User Agent used for downloads."""
        return f"ToDus {self.version_name} HTTP-Download"

    @property
    def headers_auth(self) -> Dict[str, str]:
        return {
            "Host": "auth.todus.cu",
            "User-Agent": self.auth_ua,
            "Content-Type": "application/x-protobuf",
        }

    def task_request_code(self, phone_number: str) -> None:
        headers = self.headers_auth
        data = (
            b"\n\n"
            + phone_number.encode("utf-8")
            + b"\x12\x96\x01"
            + generate_token(150).encode("utf-8")
        )
        url = "https://auth.todus.cu/v2/auth/users.reserve"
        with self.session.post(url, data=data, headers=headers) as resp:
            resp.raise_for_status()

    def request_code(self, phone_number: str) -> None:
        """Request server to send verification SMS code."""
        args = (phone_number,)
        self._run_task(self.task_request_code, self.timeout, *args)

    def task_validate_code(self, phone_number: str, code: str) -> str:
        headers = self.headers_auth
        data = (
            b"\n\n"
            + phone_number.encode("utf-8")
            + b"\x12\x96\x01"
            + generate_token(150).encode("utf-8")
            + b"\x1a\x06"
            + code.encode()
        )
        url = "https://auth.todus.cu/v2/auth/users.register"
        with self.session.post(url, data=data, headers=headers) as resp:
            resp.raise_for_status()
            if b"`" in resp.content:
                index = resp.content.index(b"`") + 1
                return resp.content[index : index + 96].decode()
            return resp.content[5:166].decode()

    def validate_code(self, phone_number: str, code: str) -> str:
        """Validate phone number with received SMS code.

        Returns the account password.
        """
        args = (
            phone_number,
            code,
        )
        return self._run_task(self.task_validate_code, self.timeout, *args)

    def task_login(self, phone_number: str, password: str) -> str:
        token = ""
        headers = self.headers_auth
        data = (
            b"\n\n"
            + phone_number.encode()
            + b"\x12\x96\x01"
            + generate_token(150).encode("utf-8")
            + b"\x12\x60"
            + password.encode()
            + b"\x1a\x05"
            + self.version_code.encode("utf-8")
        )
        url = "https://auth.todus.cu/v2/auth/token"
        with self.session.post(url, data=data, headers=headers) as resp:
            resp.raise_for_status()
            # Default Encoding for HTML4 ISO-8859-1 (Latin-1)
            token = "".join(
                c for c in resp.content.decode("latin-1") if c in string.printable
            )

        return token

    def login(self, phone_number: str, password: str) -> str:
        """Login with phone number and password to get an access token."""
        args = (
            phone_number,
            password,
        )
        return self._run_task(self.task_login, self.timeout, *args)

    def task_upload_file_1(self, token: str, size: int) -> Tuple[str, str]:
        return reserve_url(token, size)

    def task_upload_file_2(
        self,
        token: str,
        filename_path: Path,
        up_url: str,
        down_url: str,
        timeout: float,
        index: int,
    ) -> str:
        headers = {
            "User-Agent": self.upload_ua,
            "Authorization": f"Bearer {token}",
        }

        size = filename_path.stat().st_size if filename_path.exists() else 0

        with tqdm(
            total=size,
            desc=f"Part {index}",
            unit="B",
            unit_scale=True,
            unit_divisor=CHUNK_SIZE,
        ) as t:
            fileobj = open(filename_path, "rb")
            wrapped_file = CallbackIOWrapper(t.update, fileobj, "read")
            with self.session.put(
                url=up_url,
                data=wrapped_file,  # type: ignore
                headers=headers,
                timeout=timeout,
                stream=True,
            ) as resp:
                resp.raise_for_status()
            if not fileobj.closed:
                fileobj.close()
        return down_url

    def upload_file(self, token: str, filename_path: Path, index: int = 1) -> str:
        """Upload data and return the download URL."""
        size = filename_path.stat().st_size if filename_path.exists() else 0

        args = (
            token,
            size,
        )
        up_url, down_url = self._run_task(self.task_upload_file_1, self.timeout, *args)

        timeout = max(size / 1024 / 1024 * 20, self.timeout)

        args_2 = (token, filename_path, up_url, down_url, timeout, index)
        return self._run_task(self.task_upload_file_2, timeout, *args_2)

    def task_download_1(self, token: str, url: str) -> str:
        try:
            url = get_real_url(token, url)
        except socket.timeout as ex:
            logger.error(ex)
        if not url:
            raise ValueError("Invalid URL 'None'")
        return url

    def task_download_2(self, token: str, url: str, filename_path: str) -> int:
        headers = {
            "User-Agent": self.download_ua,
            "Authorization": f"Bearer {token}",
        }
        size = 0

        with self.session.get(url=url, headers=headers, stream=True) as resp:
            resp.raise_for_status()

            size = int(resp.headers.get("content-length", 0))

            file_save = Path(filename_path)

            overwrite = file_save.stat().st_size < size if file_save.exists() else True
            if overwrite:
                with tqdm.wrapattr(
                    open(file_save, "wb"),
                    "write",
                    miniters=1,
                    desc=file_save.name,
                    total=size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=CHUNK_SIZE,
                ) as fout:
                    for chunk in resp.iter_content(chunk_size=CHUNK_SIZE):
                        fout.write(chunk)
        return size

    def download_file(
        self, token: str, url: str, filename_path: str, down_timeout: float = 60 * 20
    ) -> int:
        """Download file URL.

        Returns the file size.
        """
        args = (
            token,
            url,
        )
        url = self._run_task(self.task_download_1, self.timeout, *args)
        args_2 = (
            token,
            url,
            filename_path,
        )
        return self._run_task(self.task_download_2, down_timeout, *args_2)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['todus3']

package_data = \
{'': ['*']}

install_requires = \
['multivolumefile>=0.2.3,<0.3.0',
 'py7zr>=0.16.1,<0.17.0',
 'requests>=2.25.1,<3.0.0',
 'tqdm>=4.61.2,<5.0.0']

entry_points = \
{'console_scripts': ['todus3 = todus3.main:main']}

setup_kwargs = {
    'name': 'todus3',
    'version': '0.5.4',
    'description': 'ToDus client for S3',
    'long_description': '# ToDus client for S3\n\n[![](https://img.shields.io/pypi/v/todus3.svg)](https://pypi.org/project/todus3)\n[![](https://img.shields.io/pypi/pyversions/todus3.svg)](\nhttps://pypi.org/project/todus3)\n[![](https://img.shields.io/pypi/l/todus3.svg)](https://pypi.org/project/todus3)\n[![CI](https://github.com/oleksis/todus/actions/workflows/python-ci.yml/badge.svg)](https://github.com/oleksis/todus/actions/workflows/python-ci.yml)\n[![](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nUse the ToDus API in your Python projects.\n\n## Install\n\nTo install run\n```bash\n  pip install todus3\n```\n\n## Usage\n```bash\ntodus3 -- help\n\ntodus3 -n 53123456 login\n\ntodus3 -n 53123456 download file.txt\n\ntodus3 -n 53123456 upload binary.zip -p 10485760\n```\n',
    'author': 'adbenitez',
    'author_email': 'adbenitez@nauta.cu',
    'maintainer': 'Oleksis Fraga',
    'maintainer_email': 'oleksis.fraga@gmail.com',
    'url': 'https://github.com/oleksis/todus/tree/todus3',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
